from pathlib import Path
import duckdb
import polars as pl
from src.item_matching.build_index.func_img import PipelineImage


# path
path = Path.home() / 'Downloads/cb'

# data
name = 'cb_2024-03-07.parquet'
query = f"""
select *
,concat('http://f.shopee.vn/file/', UNNEST(array_slice(string_split(images, ','), 1, 1))) image_url
from parquet_scan('{str(path / name)}')
order by item_id, images
"""
df = (
    duckdb.sql(query).pl()
    .with_columns(pl.col('create_datetime').dt.date().cast(pl.Utf8))
)
print(df.shape, df['item_name'].n_unique())


# download
path_parq = path / f'{name}_0.parquet'
df.write_parquet(path_parq)
PipelineImage(path).download_images(name)



from pathlib import Path
import polars as pl


class PipelineText:
    @staticmethod
    def clean_text(data: pl.DataFrame, col: str = 'item_name') -> pl.DataFrame:
        import re

        regex = "[\(\[\<\"].*?[\)\]\>\"]"
        return data.with_columns(
            pl.col(col).map_elements(
                lambda x: re.sub(regex, "", x).lower().rstrip('.').strip(), return_dtype=pl.String
            )
            .alias(f'{col.lower()}_clean')
        )


def ngrams_func(string: str):
    ngrams = zip(*[string[i:] for i in range(3)])
    return [''.join(ngram) for ngram in ngrams]


def tfidf(lst_item: list, dim: int = 512):
    from sklearn.feature_extraction.text import TfidfVectorizer

    vectorizer = TfidfVectorizer(analyzer=ngrams_func, max_features=dim)
    vectorizer.fit(lst_item)
    return vectorizer


def rm_all_folder(path: Path) -> None:
    for child in path.glob('*'):
        if child.is_file():
            child.unlink()
        else:
            rm_all_folder(child)
    path.rmdir()


def make_dir(folder_name: str | Path) -> None:
    """Make a directory if it doesn't exist'"""
    if isinstance(folder_name, str):
        folder_name = Path(folder_name)
    if not folder_name.exists():
        folder_name.mkdir(parents=True, exist_ok=True)



from pathlib import Path
import polars as pl
import duckdb
import sys
from loguru import logger
from tqdm import tqdm
from .func import PipelineText, make_dir

logger.remove()
logger.add(sys.stdout, colorize=True, format='<green>{time:HH:mm:ss}</green> | <level>{level}</level> | <cyan>{function}</cyan> | <level>{message}</level>')


class PipelineImage:
    def __init__(
            self,
            path: Path,
            col_image: str = 'image_url',
    ):
        self.path_image = path / 'download_img'
        self.col_image = col_image

        # init path image
        make_dir(self.path_image)

    def download_images(self, file_name: str):
        import os
        import subprocess

        os.chdir(str(self.path_image))
        command = (
            f"img2dataset --url_list={file_name}_0.parquet "
            f"--output_folder=img_{file_name}/ "
            f"--processes_count=16 "
            f"--thread_count=32 "
            f"--image_size=224 "
            f"--output_format=files "
            f"--input_format=parquet "
            f"--url_col={self.col_image} "
            f"--number_sample_per_shard=50000 "
        )
        subprocess.run(command, shell=True)

    def load_images(self, mode: str = '') -> pl.DataFrame:
        import orjson

        # listing
        path = self.path_image / f'img_{mode}'
        lst_json = sorted(path.glob('*/*.json'))
        lst_file = [orjson.loads(open(str(i), "r").read())['url'] for i in tqdm(lst_json, desc='Loading json in folder')]
        lst_img = [str(i) for i in tqdm(sorted(path.glob('*/*.jpg')), desc='Loading jpg in folder')]
        df = pl.DataFrame({
            f'{mode}_{self.col_image}': lst_file,
            f'{mode}_file_path': lst_img,
            f'{mode}_exists': [True] * len(lst_file),
        })

        logger.info(f'[Data] Load Images: {df.shape}')
        return df

    def run(
            self,
            data,
            mode: str = '',
            download: bool = False,
            edit_img_url: bool = True,
    ):
        # edit url
        col_query = f"images {self.col_image}"
        if edit_img_url:
            col_query = f",concat('http://f.shopee.vn/file/', UNNEST(array_slice(string_split(images, ','), 1, 1))) {self.col_image}"

        # load data
        query = f"""select *, {col_query} from data"""
        df = duckdb.sql(query).pl()
        logger.info(f'[Data] Base Data: {df.shape}')

        # download
        if download:
            df.write_parquet(self.path_image / f'{mode}_0.parquet')
            self.download_images(mode)

        # load data image
        data_img = self.load_images(mode)

        # join
        data = (
            df.drop(['images'])
            .pipe(PipelineText.clean_text)
            .select(pl.all().name.prefix(f'{mode}_'))
            .join(data_img, on=f'{mode}_{self.col_image}', how='left')
            .filter(pl.col(f'{mode}_exists'))
        )
        logger.info(f'[Data] Join Images: {data.shape}')
        return data, data_img



from pathlib import Path
import numpy as np
import polars as pl
from time import perf_counter
from re import search
from autofaiss import build_index
from datasets import concatenate_datasets, load_from_disk, Dataset
from .func import make_dir
from .model import Model
import sys
from loguru import logger

logger.remove()
logger.add(sys.stdout, colorize=True, format='<green>{time:HH:mm:ss}</green> | <level>{level}</level> | <cyan>{function}</cyan> | <level>{message}</level>')


class Data:
    def __init__(self, path: Path, shard_size: int = 1_500_000):
        self.path = path
        self.shard_size = shard_size

    def create_dataset(self, data: pl.DataFrame, pp, fn_kwargs: dict, col_embed: str, mode: str = '', batch_size: int = 256) -> dict[str, Path]:
        # Create directories
        path_array = self.path / f'{mode}_array'
        path_ds = self.path / f'{mode}_ds'
        path_array.mkdir(parents=True, exist_ok=True)
        path_ds.mkdir(parents=True, exist_ok=True)

        # Initialize paths
        path_tmp = {f'{mode}_array': path_array, f'{mode}_ds': path_ds}

        # Log total chunks
        total_sample = len(data)
        num_chunks = (total_sample + self.shard_size - 1) // self.shard_size
        logger.info(f'[Dataset] Total chunks {mode}: {num_chunks} - Shard size: {self.shard_size}')

        # Process and save each chunk
        for idx in range(num_chunks):
            start_idx = idx * self.shard_size
            end_idx = min(start_idx + self.shard_size, total_sample)
            dataset_chunk = Dataset.from_pandas(data[start_idx:end_idx].to_pandas())

            # Process dataset
            dataset_chunk = dataset_chunk.map(pp, batched=True, batch_size=batch_size, fn_kwargs=fn_kwargs)
            dataset_chunk.set_format(type='numpy', columns=[col_embed], output_all_columns=True)

            # Save chunk
            np.save(path_array / f'{idx}.npy', dataset_chunk[col_embed])
            dataset_chunk.save_to_disk(str(path_ds / f'{idx}'))

        return path_tmp


class BELargeScale:
    def __init__(
            self,
            path: Path,
            img_dim: bool = False,
            text_dense: bool = False,
            query_batch_size: int = 500_000
    ):
        self.img_dim = img_dim
        self.text_dense = text_dense
        self.path = path
        self.data = Data(self.path)
        self.model = Model()
        self.query_batch_size = query_batch_size

    def transform_img(
            self,
            df_db: pl.DataFrame,
            df_q: pl.DataFrame,
            col_embed: str
    ) -> dict:
        # load model
        img_model, img_processor = self.model.get_img_model()
        # transform embed
        dict_tmp = {}
        for mode, data in zip(['db', 'q'], [df_db, df_q]):
            # batch embed
            fn_kwargs = {'col': f'{mode}_file_path', 'model': img_model, 'processor': img_processor}
            path_tmp = self.data.create_dataset(
                data, mode=mode, pp=self.model.pp_img, fn_kwargs=fn_kwargs, col_embed=col_embed, batch_size=128
            )
            dict_tmp[mode] = path_tmp
        return dict_tmp

    def transform_text_dense(
            self,
            df_db: pl.DataFrame,
            df_q: pl.DataFrame,
            col_embed: str
    ) -> dict:
        # model
        text_model = self.model.get_text_model()
        # transform embed
        dict_tmp = {}
        for mode, data in zip(['db', 'q'], [df_db, df_q]):
            fn_kwargs = {'col': f'{mode}_item_name_clean', 'model': text_model}
            path_tmp = self.data.create_dataset(
                data, mode=mode, pp=self.model.pp_dense, fn_kwargs=fn_kwargs, col_embed=col_embed
            )
            dict_tmp[mode] = path_tmp
        return dict_tmp

    def match(self, df_db: pl.DataFrame, df_q: pl.DataFrame, top_k: int = 10):
        # Dataset
        col_embed = 'img_embed' if self.img_dim else 'dense_embed'
        path_tmp = {
            'db': {'db_array': self.path / 'db_array', 'db_ds': self.path / 'db_ds'},
            'q': {'q_array': self.path / 'q_array', 'q_ds': self.path / 'q_ds'},
        }
        transform_dict = {
            'img_embed': self.transform_img,
            'dense_embed': self.transform_text_dense,
        }
        if not path_tmp['db'][f'db_array'].exists():
            path_tmp = transform_dict[col_embed](df_db, df_q, col_embed)
        else:
            logger.info(f'[Matching] Dataset is existed')

        # Build index
        logger.info(f'[Matching] Start building index')
        start = perf_counter()
        path_index = self.path / 'index'
        if not path_index.exists():
            build_index(
                str(path_tmp['db']['db_array']),
                index_path=str(path_index / f'ip.index'),
                index_infos_path=str(path_index / f'index.json'),
                save_on_disk=True,
                metric_type='ip',
                verbose=30,
            )
        logger.info(f'[Matching] Building Index: {perf_counter() - start:,.2f}s')

        # Load dataset shard
        dataset_db = concatenate_datasets([
            load_from_disk(str(f)) for f in sorted(path_tmp['db']['db_ds'].glob('*'))
        ])

        # Add index
        dataset_db.load_faiss_index(col_embed, path_index / f'ip.index')

        # Dataset query shard
        dataset_q = concatenate_datasets([
            load_from_disk(str(f)) for f in sorted(path_tmp['q']['q_ds'].glob('*'))
        ])

        # Batch query
        self.path_result = self.path / 'result'
        make_dir(self.path_result)

        num_batches = len(dataset_q) // self.query_batch_size
        logger.info(f'[Matching] Start retrieve: num batches {num_batches}')
        start = perf_counter()
        for idx, i in enumerate(range(0, len(dataset_q), self.query_batch_size)):
            # init
            file_name_result = self.path_result / f'result_{idx}.parquet'
            file_name_score = self.path_result / f'score_{idx}.parquet'
            if file_name_result.exists():
                continue

            # start
            start_batch = perf_counter()
            if i + self.query_batch_size >= len(dataset_q):
                batched_queries = dataset_q[i:]
            else:
                batched_queries = dataset_q[i:i + self.query_batch_size]

            # query
            score, result = dataset_db.get_nearest_examples_batch(
                col_embed,
                batched_queries[col_embed],
                k=top_k
            )
            # export
            dict_ = {f'score_{col_embed}': [list(i) for i in score]}
            df_score = pl.DataFrame(dict_)
            df_score.write_parquet(file_name_score)
            df_result = pl.DataFrame(result).drop([col_embed])
            df_result.write_parquet(file_name_result)

            # log
            print(f"[Matching] Batch {idx}/{num_batches} match result shape: {df_result.shape} "
                  f"{perf_counter() - start_batch:,.2f}s")
            del score, result, df_score, df_result
        logger.info(f'[Matching] Retrieve: {perf_counter() - start:,.2f}s')

        # Post process
        df_score = (
            pl.concat([pl.read_parquet(f) for f in sorted(self.path_result.glob('score*.parquet'))])
        )
        df_result = (
            pl.concat([pl.read_parquet(f) for f in sorted(self.path_result.glob('result*.parquet'))])
        )
        df_match = pl.concat([df_q, df_result, df_score], how='horizontal')
        col_explode = [i for i in df_match.columns if search('db|score', i)]
        df_match = df_match.explode(col_explode)

        return df_match



import matplotlib.pyplot as plt
from PIL import Image
import polars as pl
import numpy as np


def draw_images(data, test_id):
    # search
    sample = data.filter(pl.col('q_item_id') == test_id)
    org_path = sample['q_file_path'][0]
    query_path = sample['db_file_path'].to_list()

    # plot
    fig, ax = plt.subplots(1, 1, figsize=(7, 3))
    ax.imshow(Image.open(org_path))
    fig.tight_layout()

    num_images, ncol = 5, 5
    nrow = int(np.ceil(num_images / ncol))
    fig, axes = plt.subplots(nrow, ncol, figsize=(20, 10))
    axes = axes.flatten()
    for i, image in enumerate(query_path):
        axes[i].imshow(Image.open(image))
        axes[i].axis('off')

    for _ in range(num_images, ncol * nrow, 1):
        axes[_].remove()
    fig.tight_layout()
